/*$Id: main.cu 366 2009-01-12 13:34:05Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "MarsGPU.h"

__device__ gfloat cnd(gfloat d)
{
    const gfloat       A1 = 0.31938153f;
    const gfloat       A2 = -0.356563782f;
    const gfloat       A3 = 1.781477937f;
    const gfloat       A4 = -1.821255978f;
    const gfloat       A5 = 1.330274429f;
    const gfloat RSQRT2PI = 0.39894228040143267793994605993438f;

    gfloat K = 1.0f / (1.0f + 0.2316419f * fabsf(d));

    gfloat cnd = RSQRT2PI * expf(- 0.5f * d * d) * 
        (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5)))));

    if(d > 0)
        cnd = 1.0f - cnd;

    return cnd;
}

__global__ void blackscholes(
                            gfloat* input,
                            gsize num,
                            float2* output
                            )
{
    guint32 index = blockDim.x * blockIdx.x + threadIdx.x;
    if (index >= num) return ;

    guint32 offset = index * 6;
    gfloat* buf = input + offset;

    gfloat sqrtT, expRT;
    gfloat d1, d2, CNDD1, CNDD2;

    sqrtT = sqrtf(buf[3]);
    d1 = (logf(buf[1] / buf[2]) + (buf[4] + 0.5f * buf[5] * buf[5]) * buf[3]) / (buf[5] * sqrtT);
    d2 = d1 -  buf[5] * sqrtT;

    CNDD1 = cnd(d1);
    CNDD2 = cnd(d2);

    expRT = expf(-  buf[4] *  buf[3]);
    gfloat call =  buf[1]* CNDD1 -  buf[2] * expRT * CNDD2;
    gfloat put  =  buf[2] * expRT * (1.0f - CNDD2) -  buf[1]* (1.0f - CNDD1);
    
    output[index].x = call;
    output[index].y = put;
}

int main(int argc, char** argv)
{
    if (argc < 2)
    {
        g_print("usage: %s filename\n", argv[0]);
        exit(0);
    }

    guint32 alltimer;
    cuda_start_timer(&alltimer);

    gchar* filename = argv[1];
    gchar* content = NULL;
    gsize length = 0;

    guint32 timer  ;
    cuda_start_timer(&timer);
    if (!g_file_get_contents(filename, &content, &length, NULL))
         g_error("failed to get file content!");
    cuda_stop_timer("i/o", &timer);
  
    gsize num_option = length / sizeof(gfloat) / 6;

    g_debug("num_option:%d", num_option);

    cuda_start_timer(&timer);
    guint32 i;
    gfloat* data = (gfloat*)content;
    gfloat* d_data = (gfloat*)cuda_malloc(length);
    float2* d_output = (float2*)cuda_malloc(num_option*sizeof(float2));
    cuda_memcpy_h2d(d_data, data, length);

    dim3 block_dim(128, 1, 1);
    dim3 grid_dim(1, 1, 1);

    grid_dim.x = num_option / block_dim.x + 
                (int)(num_option % block_dim.x != 0);

    g_debug("cover:%d x %d = %d", block_dim.x, grid_dim.x, 
          block_dim.x * grid_dim.x);
    blackscholes<<<grid_dim, block_dim>>>(d_data, num_option, d_output);

/*
    float2* output = (float2*)g_malloc(sizeof(float2)*num_option);
    cuda_memcpy_d2h(output, d_output, sizeof(float2)*num_option);
    for (int i = 0; i < 10; i++)
       g_debug("%f, %f", output[i].x, output[i].y); 
    g_free(output);
*/
    g_free(content);
    cuda_free((gpointer*)&d_data, length);
    cuda_free((gpointer*)&d_output, num_option*sizeof(float2));
    cuda_stop_timer("all", &alltimer);

    return 0;
}
