/*$Id: MarsGPU.cu 419 2009-02-20 07:06:47Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "MarsGPU.h"
#include "map.cu"
#include "reduce.cu"

static void mars_validate(Spec_t* spec);
#define H2D	0x01
#define D2H	0x02
static void mars_transfer(Spec_t* spec, gchar direction);
#define MAP_WORKER    0x01
#define REDUCE_WORKER    0x02
static void mars_worker(Spec_t* spec, gchar worker_type);
static void mars_group(Spec_t* spec);

/******
 main procedure
 ***/
void MarsSchedule(Spec_t* spec)
{
    guint32 timer = 0;

    cuda_start_timer(&timer);
    mars_validate(spec);
    cuda_stop_timer("validate", &timer);

    cuda_start_timer(&timer);
    mars_transfer(spec, H2D);
    cuda_stop_timer("transfer input h2d", &timer);

    cuda_start_timer(&timer);
    mars_worker(spec, MAP_WORKER);
    cuda_stop_timer("map-total", &timer);

    if (spec->workflow != MAP_ONLY)
    {
        cuda_start_timer(&timer);
        mars_group(spec);
        cuda_stop_timer("group", &timer);

        if (spec->workflow == MAP_GROUP_REDUCE)
        {
          int2* group_index = (int2*)cuda_malloc(spec->_inter_group_index_size);
          cuda_memcpy_h2d(group_index, spec->_inter_group_index, spec->_inter_group_index_size);
          cuda_free_host((gpointer*)&spec->_inter_group_index, spec->_inter_group_index_size); 
          spec->_inter_group_index = group_index;
 
          cuda_start_timer(&timer);
          mars_worker(spec, REDUCE_WORKER);
          cuda_stop_timer("reduce", &timer);

          if (spec->attr & OUTPUT_TO_HOST)
          {
              gpointer keys = cuda_malloc_host(spec->_output_keys_size);
              gpointer vals = cuda_malloc_host(spec->_output_vals_size);
              int4* index = (int4*)cuda_malloc_host(spec->_output_index_size);

              cuda_memcpy_d2h(keys, spec->_output_keys, spec->_output_keys_size);
              cuda_memcpy_d2h(vals, spec->_output_vals, spec->_output_vals_size);
              cuda_memcpy_d2h(index, spec->_output_index, spec->_output_index_size);

              cuda_free((gpointer*)&spec->_output_keys, spec->_output_keys_size);
              cuda_free((gpointer*)&spec->_output_vals, spec->_output_vals_size);
              cuda_free((gpointer*)&spec->_output_index, spec->_output_index_size);

              spec->_output_keys = keys;
              spec->_output_vals = vals;
              spec->_output_index = index;
 
              return;
          }
          else
          {
              return;
          }
        }
        else if (spec->attr & OUTPUT_TO_HOST)
        {
            spec->_output_num = spec->_inter_num;
            spec->_output_keys_size = spec->_inter_keys_size;
            spec->_output_vals_size = spec->_inter_vals_size;
            spec->_output_index_size = spec->_inter_index_size;
            spec->_output_group_index_size = spec->_inter_group_index_size;
            spec->_output_group_num = spec->_inter_group_num;

            spec->_output_keys = cuda_malloc_host(spec->_output_keys_size);
            spec->_output_vals = cuda_malloc_host(spec->_output_vals_size);
            spec->_output_index = (int4*)cuda_malloc_host(spec->_output_index_size);
            spec->_output_group_index = spec->_inter_group_index;

            cuda_memcpy_d2h(spec->_output_keys, spec->_inter_keys, spec->_output_keys_size);
            cuda_memcpy_d2h(spec->_output_vals, spec->_inter_vals, spec->_output_vals_size);
            cuda_memcpy_d2h(spec->_output_index, spec->_inter_index, spec->_output_index_size);

            cuda_free((gpointer*)&spec->_inter_keys, spec->_inter_keys_size);
            cuda_free((gpointer*)&spec->_inter_vals, spec->_inter_vals_size);
            cuda_free((gpointer*)&spec->_inter_index, spec->_inter_index_size);
            cuda_free((gpointer*)&spec->_inter_group_index, spec->_inter_group_index_size);
 
        
            spec->_output_in_gpu = 0;
 
            return;
        }
        else
        {
            spec->_output_group_index_size = spec->_inter_group_index_size;
            spec->_output_group_num = spec->_inter_group_num;
            spec->_output_group_index = (int2*)cuda_malloc(spec->_inter_group_index_size);
            cuda_memcpy_h2d(spec->_output_group_index, spec->_inter_group_index, spec->_inter_group_index_size);
        }
    }
    else if (spec->attr & OUTPUT_TO_HOST)
    {
        spec->_output_num = spec->_inter_num;
        spec->_output_keys_size = spec->_inter_keys_size;
        spec->_output_vals_size = spec->_inter_vals_size;
        spec->_output_index_size = spec->_inter_index_size;

        spec->_output_keys = cuda_malloc_host(spec->_output_keys_size);
        spec->_output_vals = cuda_malloc_host(spec->_output_vals_size);
        spec->_output_index = (int4*)cuda_malloc_host(spec->_output_index_size);

        cuda_memcpy_d2h(spec->_output_keys, spec->_inter_keys, spec->_output_keys_size);
        cuda_memcpy_d2h(spec->_output_vals, spec->_inter_vals, spec->_output_vals_size);
        cuda_memcpy_d2h(spec->_output_index, spec->_inter_index, spec->_output_index_size);

        cuda_free((gpointer*)&spec->_inter_keys, spec->_inter_keys_size);
        cuda_free((gpointer*)&spec->_inter_vals, spec->_inter_vals_size);
        cuda_free((gpointer*)&spec->_inter_index, spec->_inter_index_size);
 
        
        spec->_output_in_gpu = 0;
        return;
    }
    spec->_output_keys = spec->_inter_keys;
    spec->_output_vals = spec->_inter_vals;
    spec->_output_index = spec->_inter_index;
    spec->_output_num = spec->_inter_num;
    spec->_output_keys_size = spec->_inter_keys_size;
    spec->_output_vals_size = spec->_inter_vals_size;
    spec->_output_index_size = spec->_inter_index_size;
    spec->_output_in_gpu = 1;
}

/******
 Setup and Teardown 
 ***/
Spec_t* MarsSpecNew()
{
    Spec_t* spec = (Spec_t*)cuda_malloc_host(sizeof(Spec_t));

    //Get hardware infomation
    struct cudaDeviceProp cdp;
    cudaGetDeviceProperties(&cdp, 0);
    spec->_max_size_dmem = (gsize)cdp.totalGlobalMem;
    spec->_max_num_reg_per_block = (gint32)cdp.regsPerBlock;
    spec->_max_size_smem_per_block = (gsize)cdp.sharedMemPerBlock;
    spec->_size_warp = (gint32)cdp.warpSize;
    spec->_max_thread_per_block = (gint32)cdp.maxThreadsPerBlock;
    spec->_ver = cdp.minor;

    return spec;
}

void MarsFinish(Spec_t** spec)
{
    Spec_t* gspec = *spec;

    if (gspec->_output_in_gpu)
    {
        cuda_free((gpointer*)&gspec->_output_keys, gspec->_output_keys_size);
        cuda_free((gpointer*)&gspec->_output_vals, gspec->_output_vals_size);
        cuda_free((gpointer*)&gspec->_output_index, gspec->_output_index_size);
        if (gspec->workflow == MAP_GROUP)
            cuda_free((gpointer*)&gspec->_output_group_index, gspec->_output_group_index_size);
    }
    else
    {
        cuda_free_host((gpointer*)&gspec->_output_keys, gspec->_output_keys_size);
        cuda_free_host((gpointer*)&gspec->_output_vals, gspec->_output_vals_size);
        cuda_free_host((gpointer*)&gspec->_output_index, gspec->_output_index_size);
        if (gspec->workflow == MAP_GROUP)
            cuda_free_host((gpointer*)&gspec->_output_group_index, gspec->_output_group_index_size);
 
    }

    cuda_free_host((gpointer*)spec, sizeof(Spec_t));
}

void MarsAddInputRec(Spec_t* spec, gpointer key, gsize keySize, gpointer val, gsize valSize)
{
    static int2 curOffset;
    static int3 curChunkNum;
    const size_t dataChunkSize = 1024*1024*200;
    int index = spec->_input_num;

    if (spec->_input_num > 0)
    {
        if (dataChunkSize*curChunkNum.x < (curOffset.x + keySize))
        {
            gsize newsize = spec->_input_keys_buf_size;
            gsize oldsize = newsize - dataChunkSize;
            spec->_input_keys = (gpointer)cuda_realloc_host(spec->_input_keys, oldsize, newsize);
            ++curChunkNum.x;
            spec->_input_keys_buf_size += dataChunkSize;
        }

        if (dataChunkSize*curChunkNum.y < (curOffset.y + valSize))
        {
            gsize newsize = spec->_input_vals_buf_size;
            gsize oldsize = newsize - dataChunkSize;
            spec->_input_vals = (gpointer)cuda_realloc_host(spec->_input_vals, oldsize, newsize);
            ++curChunkNum.y;
            spec->_input_vals_buf_size += dataChunkSize;
        }

        if (dataChunkSize*curChunkNum.z < (spec->_input_num+1)*sizeof(int4))
        {
            gsize newsize = spec->_input_index_buf_size;
            gsize oldsize = newsize - dataChunkSize;
            spec->_input_index = (int4*)cuda_realloc_host(spec->_input_index, oldsize, newsize);
            ++curChunkNum.z;
            spec->_input_index_buf_size += dataChunkSize;
        }
    }
    else
    {
	spec->_input_keys = (gpointer)g_malloc(dataChunkSize);
	spec->_input_vals = (gpointer)g_malloc(dataChunkSize);
	spec->_input_index = (int4*)g_malloc(dataChunkSize);

        spec->_input_keys_buf_size = dataChunkSize;
        spec->_input_vals_buf_size = dataChunkSize;
        spec->_input_index_buf_size = dataChunkSize;

	curChunkNum.x++;
	curChunkNum.y++;
	curChunkNum.z++;
    }

    memcpy((char*)spec->_input_keys+curOffset.x, key, keySize);
    memcpy((char*)spec->_input_vals+curOffset.y, val, valSize);	
    spec->_input_index[index].x = curOffset.x;
    spec->_input_index[index].y = keySize;
    spec->_input_index[index].z = curOffset.y;
    spec->_input_index[index].w = valSize;

    curOffset.x += keySize;
    curOffset.y += valSize;
    spec->_input_keys_size += keySize;
    spec->_input_vals_size += valSize;
    spec->_input_index_size += sizeof(int4);

    ++spec->_input_num;
}

/******
 Debug functions
 ***/
void MarsPrintGPUInfo(Spec_t* spec)
{
    g_message("====== GPU Info ======");
    g_message("maximum global memory size: %d MB", 
           spec->_max_size_dmem / 1024 / 1024);
    g_message("maximum number of registers per block: %d",
           spec->_max_num_reg_per_block);
    g_message("maximum size of shared memory per block: %d KB",
           spec->_max_size_smem_per_block/1024);
    g_message("warp size: %d\n", spec->_size_warp);
    g_message("maximum size of threads per block: %d", 
           spec->_max_thread_per_block);
    g_message("compatibility version: 1.%d", 
           spec->_ver);
}

void MarsProcessRec(Spec_t* spec, gsize max_num, ProcessFunc_t func, gpointer arg)
{ 
    gpointer keys = NULL; 
    gpointer vals = NULL; 
    int4* index = NULL; 
    gsize rec_num = 0;

    keys = spec->_output_keys; 
    vals = spec->_output_vals; 
    index = spec->_output_index; 
    rec_num = spec->_output_num; 

    if (max_num <=0 || max_num >= rec_num) max_num = rec_num;

    for (int i = 0; i < max_num; i++)
    {
       gpointer key = (char*)keys + index[i].x;
       gpointer val = (char*)vals + index[i].z;
       func(key, index[i].y, val, index[i].w, i, arg); 
    }
}

void MarsProcessGroup(Spec_t* spec, gchar type, gsize max_num, ProcessGroupFunc_t func, gpointer arg)
{
    gpointer keys = NULL; 
    gpointer vals = NULL; 
    int4* index = NULL; 
    int2* group_index = NULL; 
    gsize group_num = 0;

    keys = spec->_output_keys; 
    vals = spec->_output_vals; 
    index = spec->_output_index; 
    group_index = spec->_output_group_index;
    group_num = spec->_output_group_num;

    if (max_num <=0 || max_num >= group_num) max_num = group_num;

    for (int i = 0; i < max_num; i++)
    {
        gpointer key = (gchar*)keys + index[group_index[i].x].x; 
        gsize valnum = group_index[i].y - group_index[i].x; 

        func(key, index[group_index[i].x].y, valnum, spec, i, arg);
    }
}

gpointer GetVal(Spec_t* spec, gsize group_id, gsize rec_id)
{
    return (gpointer)((gchar*)spec->_output_vals + 
spec->_output_index[spec->_output_group_index[group_id].x + rec_id].z); 
}

__device__ gpointer GetReduceVal(gpointer vals, int4* index, int2* group_index, gsize group_num, gsize rec_num)
{
   gint i = group_index[group_num].x + rec_num;
   int4 rindex = index[i];
   return (gchar*)vals + rindex.z;
}

/******
 ***/

static void mars_validate(Spec_t* spec)
{
    if (spec->workflow != MAP_ONLY &&
        spec->workflow != MAP_GROUP &&
        spec->workflow != MAP_GROUP_REDUCE)
        spec->workflow = MAP_ONLY;


    if (spec->warp_num <= 0 ||
        spec->warp_num > (spec->_max_thread_per_block / spec->_size_warp))
        spec->warp_num = 4;

}

static void mars_transfer(Spec_t* spec, gchar direction)
{
    if (spec->_input_in_gpu) return;
    gpointer keys = NULL; 
    gpointer vals = NULL; 
    int4* index = NULL; 
    gsize keys_size = 0;
    gsize vals_size = 0; 
    gsize index_size = 0; 
    gchar* in_gpu = NULL; 

    keys = spec->_input_keys; 
    vals = spec->_input_vals; 
    index = spec->_input_index; 
    keys_size = spec->_input_keys_size; 
    vals_size = spec->_input_vals_size; 
    index_size = spec->_input_index_size; 
    in_gpu = &spec->_input_in_gpu; 

    if (direction == H2D)
    {
        gpointer tmpKeys = cuda_malloc(keys_size);
        gpointer tmpVals = cuda_malloc(vals_size);
        int4* tmpIndex = (int4*)cuda_malloc(index_size);

        cuda_memcpy_h2d(tmpKeys, keys, keys_size);
        cuda_memcpy_h2d(tmpVals, vals, vals_size);
        cuda_memcpy_h2d(tmpIndex, index, index_size);

        *in_gpu = 1;

        g_free(keys);
        g_free(vals);
        g_free(index);

        spec->_input_keys = tmpKeys;
        spec->_input_vals = tmpVals;
        spec->_input_index = tmpIndex;
    }
    else  if (direction == D2H)
    {
    }
}


__device__ void emit_inter_count(gsize keySize,
                               gsize valSize,
                               guint32* keyHist,
                               guint32* valHist,
                               guint32* indexHist)
{
    guint32 index = (blockIdx.y * blockDim.y + blockIdx.x) * blockDim.x + threadIdx.x;
    keyHist[index] += keySize;
    valHist[index] += valSize;
    ++indexHist[index];
}

typedef char4 vector_t;
__device__ void emit_inter(
                         gpointer key,
                         gsize keySize,
                         gpointer val,
                         gsize valSize,
                         guint32* ps_keyoffset,
                         guint32* ps_valoffset,
                         guint32* ps_count,
                         int2* keyval_offset,
                         guint32* cur_index,
                         gpointer output_keys,
                         gpointer output_vals,
                         int4* output_index
                          )
{
	guint32 index = (blockIdx.y * blockDim.y + blockIdx.x) * blockDim.x + threadIdx.x;

	gchar* pkey = (gchar*)output_keys + ps_keyoffset[index] + keyval_offset[index].x;
	gchar* pval = (gchar*)output_vals + ps_valoffset[index] + keyval_offset[index].y;

        gchar* skey = (gchar*)key; 
        gchar* sval = (gchar*)val; 

        guint32 i;
#ifndef __VECTOR__
        for (i = 0; i < keySize; i++)
            pkey[i] = skey[i];
        for (i = 0; i < valSize; i++)
            pval[i] = sval[i];
#else
        guint32 vector4Size, vector4remain;
        vector_t* p4key = (vector_t*)pkey;
        vector_t* p4val = (vector_t*)pval;
        vector_t* s4key = (vector_t*)skey;
        vector_t* s4val = (vector_t*)sval;

        vector4Size = keySize / sizeof(vector_t);
        for (i = 0; i < vector4Size; i++)
            p4key[i] = s4key[i];
        vector4remain = keySize % sizeof(vector_t);
        for (i = 0; i < vector4remain; i++)
            p4key[i] = s4key[i]; 

        vector4Size = valSize / sizeof(vector_t);
        for (i = 0; i < vector4Size; i++)
            p4val[i] = s4val[i];
        vector4remain = valSize % sizeof(vector_t);
        for (i = 0; i < vector4remain; i++)
            p4val[i] = s4val[i]; 
#endif

	keyval_offset[index].x += keySize;
	keyval_offset[index].y += valSize;

	if (cur_index[index] != 0)
	{
	output_index[ps_count[index] + cur_index[index]].x = 
		(output_index[ps_count[index] + cur_index[index] - 1].x + 
		 output_index[ps_count[index] + cur_index[index] - 1].y);
	output_index[ps_count[index] + cur_index[index]].z = 
		(output_index[ps_count[index] + cur_index[index] - 1].z + 
		 output_index[ps_count[index] + cur_index[index] - 1].w);
	}
	
	output_index[ps_count[index] + cur_index[index]].y = keySize;
	output_index[ps_count[index] + cur_index[index]].w = valSize;

	++cur_index[index];
}


__global__ void map_count(
                         gpointer input_keys,
                         gpointer input_vals,
                         int4*     input_index,
                         guint32*  key_hist,
                         guint32*  val_hist,
                         guint32*  index_hist,
                         gsize     rec_num,
                         guint32   rec_per_thread 
                         )
{
    //for each record
    for (guint32 i = 0; i < rec_per_thread; i++)
    {
        //coalesced access to global memory
        guint32 block_offset = (blockIdx.y *gridDim.y + blockIdx.x) * rec_per_thread*blockDim.x;
        guint32 rec_offset =  i * blockDim.x;
        guint32 thread_offset = threadIdx.x;
        guint32 rec_index = block_offset + rec_offset + thread_offset; 
        if (rec_index >= rec_num) break;


        int4 index = input_index[rec_index];

        gpointer key = (char*)input_keys + index.x;
        gpointer val = (char*)input_vals + index.z;
 
       map_count_func(
                      key,
                      val,
                      index.y,
                      index.w,
                      key_hist,
                      val_hist,
                      index_hist 
                      );
    }
}

__global__ void map(
                    gpointer input_keys,
                    gpointer input_vals,
                    int4*    input_index,
                    guint32* ps_keyoffset,
                    guint32* ps_valoffset,
                    guint32* ps_count,
                    int2*    keyval_offset,
                    guint32*   cur_index,
                    gsize    rec_num,
                    gsize    rec_per_thread,
                    gpointer output_keys,
                    gpointer output_vals,
                    int4*    output_index
                   )
{
    guint32 idx = (blockDim.y*blockIdx.y + blockIdx.x) * blockDim.x + threadIdx.x;
    output_index[ps_count[idx]].x = ps_keyoffset[idx];
    output_index[ps_count[idx]].z = ps_valoffset[idx];

    //for each record
    for (guint32 i = 0; i < rec_per_thread; i++)
    {
        //coalesced access to global memory
        guint32 block_offset = (blockIdx.y *gridDim.y + blockIdx.x) * rec_per_thread*blockDim.x;
        guint32 rec_offset =  i * blockDim.x;
        guint32 thread_offset = threadIdx.x;
        guint32 rec_index = block_offset + rec_offset + thread_offset; 
        if (rec_index >= rec_num) break;

        int4 index = input_index[rec_index];

        gpointer key = (char*)input_keys + index.x;
        gpointer val = (char*)input_vals + index.z;
 
        map_func(
                key,
                val,
                index.y,
                index.w,
                ps_keyoffset,
                ps_valoffset,
                ps_count,
                keyval_offset,
                cur_index,
                output_keys,
                output_vals,
                output_index
                );
    }
}

__global__ void reduce_count(
                         gpointer input_keys,
                         gpointer input_vals,
                         int4*     input_index,
                         guint32*  key_hist,
                         guint32*  val_hist,
                         guint32*  index_hist,
                         gsize     rec_num,
                         guint32   rec_per_thread ,
                         int2*   group_index
                         )
{
    //for each record
    for (guint32 i = 0; i < rec_per_thread; i++)
    {
        //coalesced access to global memory
        guint32 block_offset = (blockIdx.y *gridDim.y + blockIdx.x) * rec_per_thread*blockDim.x;
        guint32 rec_offset =  i * blockDim.x;
        guint32 thread_offset = threadIdx.x;
        guint32 rec_index = block_offset + rec_offset + thread_offset; 
        if (rec_index >= rec_num) break;

        int2 the_group_index = group_index[rec_index];
        int4 index = input_index[the_group_index.x];
        gsize valNum = the_group_index.y - the_group_index.x;

        gpointer key = (char*)input_keys + index.x;
        gpointer val = input_vals;
 
       reduce_count_func(
                      key,
                      val,
                      index.y,
                      valNum,
                      key_hist,
                      val_hist,
                      index_hist, 
                      input_index,
                      group_index,
                      rec_index
                      );
    }
}

__global__ void reduce(
                    gpointer input_keys,
                    gpointer input_vals,
                    int4*    input_index,
                    guint32* ps_keyoffset,
                    guint32* ps_valoffset,
                    guint32* ps_count,
                    int2*    keyval_offset,
                    guint32*   cur_index,
                    gsize    rec_num,
                    gsize    rec_per_thread,
                    gpointer output_keys,
                    gpointer output_vals,
                    int4*    output_index,
                    int2*    group_index
                   )
{
    guint32 idx = (blockDim.y*blockIdx.y + blockIdx.x) * blockDim.x + threadIdx.x;
    output_index[ps_count[idx]].x = ps_keyoffset[idx];
    output_index[ps_count[idx]].z = ps_valoffset[idx];

    //for each record
    for (guint32 i = 0; i < rec_per_thread; i++)
    {
        //coalesced access to global memory
        guint32 block_offset = (blockIdx.y *gridDim.y + blockIdx.x) * rec_per_thread*blockDim.x;
        guint32 rec_offset =  i * blockDim.x;
        guint32 thread_offset = threadIdx.x;
        guint32 rec_index = block_offset + rec_offset + thread_offset; 
        if (rec_index >= rec_num) break;

        int2 the_group_index = group_index[rec_index];
        int4 index = input_index[the_group_index.x];
        gsize valNum = the_group_index.y - the_group_index.x;


        gpointer key = (char*)input_keys + index.x;
        gpointer val = input_vals;
 
        reduce_func(
                key,
                val,
                index.y,
                valNum,
                ps_keyoffset,
                ps_valoffset,
                ps_count,
                keyval_offset,
                cur_index,
                output_keys,
                output_vals,
                output_index,
                input_index,
                group_index,
                rec_index
                );
    }
}
static void mars_worker(Spec_t* spec, gchar worker_type)
{
    gpointer input_keys = NULL;
    gpointer input_vals = NULL;
    int4* input_index = NULL;
    int2* group_index = NULL;
    gsize rec_num = 0;
    gsize group_num = 0;
    guint32 rec_per_thread = 0;
    gsize input_keysize = 0;
    gsize input_valsize = 0;
    gsize input_indexsize = 0;

    if (worker_type == MAP_WORKER)
    {
        input_keys = spec->_input_keys;
        input_vals = spec->_input_vals;
        input_index = spec->_input_index;
        rec_num = spec->_input_num;
        rec_per_thread = 1;//spec->rec_per_map_task;
        input_keysize = spec->_input_keys_size;
        input_valsize = spec->_input_vals_size;
        input_indexsize = spec->_input_index_size;
    }
    else
    {
        input_keys = spec->_inter_keys;
        input_vals = spec->_inter_vals;
        input_index = spec->_inter_index;
        rec_num = spec->_inter_group_num;
        //group_num = spec->_inter_group_num;
        rec_per_thread = 1;//spec->rec_per_reduce_task;
        group_index = spec->_inter_group_index;
        input_keysize = spec->_inter_keys_size;
        input_valsize = spec->_inter_vals_size;
        input_indexsize = spec->_inter_index_size;
    }

    if (rec_num <= 0)
        g_error("No input data!");

    dim3 thread_dim(spec->warp_num * spec->_size_warp,1,1);
    guint32 block_num = rec_num / (thread_dim.x*rec_per_thread) + 
         (guint32)(rec_num % (thread_dim.x*rec_per_thread) != 0);
    dim3 block_dim(block_num,1);
    if (block_num > 65535)
    {
        block_dim.x = sqrt(block_num); 
        block_dim.y = block_num / block_dim.x +
              (guint32)(block_num % block_dim.x != 0);
    }
    guint32 total_num_thread = thread_dim.x * block_dim.x * block_dim.y;

    g_debug("num_rec:%d", rec_num);
    g_debug("thread_num:%d", thread_dim.x);
    g_debug("block_num:%d", block_num);
    g_debug("block_dim:%d x %d = %d", block_dim.x, block_dim.y, block_dim.y * block_dim.x);
    g_debug("total_num_thread: %d", total_num_thread);

    //mapcount
    guint32 hist_size = sizeof(guint32)*total_num_thread;
    guint32* key_hist = (guint32*)cuda_malloc(hist_size);
    guint32* val_hist = (guint32*)cuda_malloc(hist_size);
    guint32* index_hist = (guint32*)cuda_malloc(hist_size);

    if (worker_type == MAP_WORKER)
    {
    map_count<<<block_dim, thread_dim>>>(
                                        input_keys,
                                        input_vals,
                                        input_index,
                                        key_hist,
                                        val_hist,
                                        index_hist,
                                        rec_num, rec_per_thread
                                        );
    }
    else
    {
    reduce_count<<<block_dim, thread_dim>>>(
                                        input_keys,
                                        input_vals,
                                        input_index,
                                        key_hist,
                                        val_hist,
                                        index_hist,
                                        rec_num,
                                        rec_per_thread,
                                        group_index
                                        );
    }
    cuda_sync();

    //scan
    guint32* ps_keyoffset = (guint32*)cuda_malloc(hist_size);
    guint32* ps_valoffset = (guint32*)cuda_malloc(hist_size);
    guint32* ps_count = (guint32*)cuda_malloc(hist_size);

    guint32 output_keysize = prefixSum(key_hist, ps_keyoffset, total_num_thread);
    guint32 output_valsize = prefixSum(val_hist, ps_valoffset, total_num_thread);
    guint32 output_recnum = prefixSum(index_hist, ps_count, total_num_thread);

    cuda_free((gpointer*)&key_hist, hist_size);
    cuda_free((gpointer*)&val_hist, hist_size);
    cuda_free((gpointer*)&index_hist, hist_size);

    //map or reduce
    int2* keyval_offset = (int2*)cuda_malloc(sizeof(int2)*total_num_thread); 
    guint32* cur_index = (guint32*)cuda_malloc(sizeof(guint32)*total_num_thread);
    
    gpointer output_keys = cuda_malloc(output_keysize);
    gpointer output_vals = cuda_malloc(output_valsize);
    int4* output_index = (int4*)cuda_malloc(output_recnum*sizeof(int4));

    if (worker_type == MAP_WORKER)
    {
    map<<<block_dim, thread_dim>>>(
                                  input_keys,
                                  input_vals,
                                  input_index,
                                  ps_keyoffset,
                                  ps_valoffset,
                                  ps_count,
                                  keyval_offset,
                                  cur_index,
                                  rec_num,
                                  rec_per_thread,
                                  output_keys,
                                  output_vals,
                                  output_index
                                  ); 
    }
    else
    {
    reduce<<<block_dim, thread_dim>>>(
                                  input_keys,
                                  input_vals,
                                  input_index,
                                  ps_keyoffset,
                                  ps_valoffset,
                                  ps_count,
                                  keyval_offset,
                                  cur_index,
                                  rec_num,
                                  rec_per_thread,
                                  output_keys,
                                  output_vals,
                                  output_index,
                                  group_index
                                  ); 
 
    }
    cuda_sync();

    cuda_free((gpointer*)&keyval_offset, sizeof(int2)*total_num_thread);
    cuda_free((gpointer*)&cur_index, sizeof(gsize)*total_num_thread);

    cuda_free((gpointer*)&ps_keyoffset, hist_size);
    cuda_free((gpointer*)&ps_valoffset, hist_size);
    cuda_free((gpointer*)&ps_count, hist_size);

    cuda_free((gpointer*)&input_keys, input_keysize);
    cuda_free((gpointer*)&input_vals, input_valsize);
    cuda_free((gpointer*)&input_index, input_indexsize);

    if (worker_type == MAP_WORKER)
    {
    spec->_inter_keys = output_keys;
    spec->_inter_vals = output_vals;
    spec->_inter_index = output_index;
    spec->_inter_num = output_recnum;
    spec->_inter_keys_size = output_keysize;
    spec->_inter_vals_size = output_valsize;
    spec->_inter_index_size = output_recnum*sizeof(int4);
    spec->_inter_in_gpu = 1;
    }
    else
    {
    cuda_free((gpointer*)&group_index, sizeof(int2)*group_num);

    spec->_output_keys = output_keys;
    spec->_output_vals = output_vals;
    spec->_output_index = output_index;
    spec->_output_num = output_recnum;
    spec->_output_keys_size = output_keysize;
    spec->_output_vals_size = output_valsize;
    spec->_output_index_size = output_recnum*sizeof(int4);
    spec->_output_in_gpu = 1;
 
    }
}

static void mars_group(Spec_t* spec)
{
    gpointer output_keys = cuda_malloc(spec->_inter_keys_size);
    gpointer output_vals = cuda_malloc(spec->_inter_vals_size);
    int4* output_index = (int4*)cuda_malloc(spec->_inter_index_size);
    int2* output_group_index = NULL;
    gsize group_num = 0;

    saven_initialPrefixSum(spec->_inter_num);
    group_num = sort_GPU(
                        spec->_inter_keys,
                        spec->_inter_keys_size,
                        spec->_inter_vals,
                        spec->_inter_vals_size,
                        spec->_inter_index,
                        spec->_inter_num,
                        output_keys,
                        output_vals,
                        output_index,
                        &output_group_index
                        );

    cuda_free(&spec->_inter_keys, spec->_inter_keys_size);
    cuda_free(&spec->_inter_vals, spec->_inter_vals_size);
    cuda_free((gpointer*)&spec->_inter_index, spec->_inter_index_size);

    spec->_inter_keys = output_keys;
    spec->_inter_vals = output_vals;
    spec->_inter_index = output_index;
    spec->_inter_group_index = output_group_index;
    spec->_inter_group_num = group_num;
    spec->_inter_group_index_size = sizeof(int2)*group_num;

    g_debug("group_num:%d", group_num);
}

