/*$Id: MarsCPU.c 375 2009-01-14 11:23:36Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "MarsCPU.h"
#include "map.c"
#include "reduce.c"

static void mars_validate(Spec_t* spec);
#define MAP_WORKER       0x00
#define REDUCE_WORKER    0x01
static void mars_worker(Spec_t* spec, gchar type);
static void mars_group(Spec_t* spec);

/******
 main procedure
 ***/
void MarsSchedule(Spec_t* spec)
{
    mars_validate(spec);

    if (!g_thread_supported())
       g_thread_init(NULL);
 
    Timer_t timer;
    cpu_start_timer(&timer);
    mars_worker(spec, MAP_WORKER);
    cpu_stop_timer("map", &timer);

    if (spec->workflow != MAP_ONLY)
    {
        cpu_start_timer(&timer);
        mars_group(spec);
        cpu_stop_timer("group", &timer);

        if (spec->workflow == MAP_GROUP_REDUCE)
        {
            cpu_start_timer(&timer);
            mars_worker(spec, REDUCE_WORKER);
            cpu_stop_timer("reduce", &timer);

            return;
        }
        else
        {
            spec->_output_group_index = spec->_inter_group_index;
            spec->_output_group_index_size = spec->_inter_group_index_size;
            spec->_output_group_num = spec->_inter_group_num;
        }
    }

    spec->_output_keys = spec->_inter_keys;
    spec->_output_vals = spec->_inter_vals;
    spec->_output_index = spec->_inter_index;
    spec->_output_num = spec->_inter_num;
    spec->_output_keys_size = spec->_inter_keys_size;
    spec->_output_vals_size = spec->_inter_vals_size;
    spec->_output_index_size = spec->_inter_index_size;
}


/******
 Setup and Teardown 
 ***/
Spec_t* MarsSpecNew()
{
    Spec_t* spec = g_new0(Spec_t, 1);

    cpu_set_t cpus;
    __CPU_ZERO(&cpus);
    int i;
    sched_getaffinity(0, sizeof(cpus), &cpus);
    for (i = 0; i < sizeof(cpus); i++)
    {
        spec->_proc_num += __CPU_ISSET(i, &cpus);
    }

    //g_debug("# of processors:%d", spec->_proc_num);
    return spec;
}

void MarsFinish(Spec_t** spec)
{
    Spec_t* pspec = *spec;

    g_free(pspec->_output_keys);
    g_free(pspec->_output_vals);
    g_free(pspec->_output_index);

    if (pspec->workflow == MAP_GROUP)
        g_free(pspec->_output_group_index);

    g_free(*spec);
    *spec = NULL;
}

void MarsAddInputRec(Spec_t* spec, gpointer key, gsize keySize, gpointer val, gsize valSize)
{
    static int2 curOffset;
    static int3 curChunkNum;
    const size_t dataChunkSize = 1024*1024*200;
    int index = spec->_input_num;

    if (spec->_input_num > 0)
    {
        if (dataChunkSize*curChunkNum.x < (curOffset.x + keySize))
        {
            gsize newsize = spec->_input_keys_buf_size;
            spec->_input_keys = (gpointer)g_realloc(spec->_input_keys, newsize);
            ++curChunkNum.x;
            spec->_input_keys_buf_size += dataChunkSize;
        }

        if (dataChunkSize*curChunkNum.y < (curOffset.y + valSize))
        {
            gsize newsize = spec->_input_vals_buf_size;
            spec->_input_vals = (gpointer)g_realloc(spec->_input_vals, newsize);
            ++curChunkNum.y;
            spec->_input_vals_buf_size += dataChunkSize;
        }

        if (dataChunkSize*curChunkNum.z < (spec->_input_num+1)*sizeof(int4))
        {
            gsize newsize = spec->_input_index_buf_size;
            spec->_input_index = (int4*)g_realloc(spec->_input_index, newsize);
            ++curChunkNum.z;
            spec->_input_index_buf_size += dataChunkSize;
        }
    }
    else
    {
	spec->_input_keys = (gpointer)g_malloc0(dataChunkSize);
	spec->_input_vals = (gpointer)g_malloc0(dataChunkSize);
	spec->_input_index = (int4*)g_malloc0(dataChunkSize);

        spec->_input_keys_buf_size = dataChunkSize;
        spec->_input_vals_buf_size = dataChunkSize;
        spec->_input_index_buf_size = dataChunkSize;

	curChunkNum.x++;
	curChunkNum.y++;
	curChunkNum.z++;
    }

    memcpy((char*)spec->_input_keys+curOffset.x, key, keySize);
    memcpy((char*)spec->_input_vals+curOffset.y, val, valSize);	
    spec->_input_index[index].x = curOffset.x;
    spec->_input_index[index].y = keySize;
    spec->_input_index[index].z = curOffset.y;
    spec->_input_index[index].w = valSize;

    curOffset.x += keySize;
    curOffset.y += valSize;
    spec->_input_keys_size += keySize;
    spec->_input_vals_size += valSize;
    spec->_input_index_size += sizeof(int4);

    ++spec->_input_num;
}

/******
 Debug functions
 ***/
#define   FETCH_BUFFER do{\
    switch(type) \
    { \
        case IN_REC: \
            keys = spec->_input_keys; \
            vals = spec->_input_vals; \
            index = spec->_input_index; \
            keys_size = spec->_input_keys_size; \
            vals_size = spec->_input_vals_size; \
            index_size = spec->_input_index_size; \
            rec_num = spec->_input_num; \
            break; \
        case INTER_REC: \
            keys = spec->_inter_keys; \
            vals = spec->_inter_vals; \
            index = spec->_inter_index; \
            keys_size = spec->_inter_keys_size; \
            vals_size = spec->_inter_vals_size; \
            index_size = spec->_inter_index_size; \
            rec_num = spec->_inter_num; \
            group_index = spec->_inter_group_index;\
            group_num = spec->_inter_group_num;\
            break; \
        case OUT_REC: \
            keys = spec->_output_keys; \
            vals = spec->_output_vals; \
            index = spec->_output_index; \
            keys_size = spec->_output_keys_size; \
            vals_size = spec->_output_vals_size; \
            index_size = spec->_output_index_size; \
            rec_num = spec->_output_num; \
            group_index = spec->_output_group_index;\
            group_num = spec->_output_group_num;\
            break; \
        default: \
            return; \
    } \
}while(0) 


void MarsPrintRecInfo(Spec_t* spec, gchar type)
{
    gpointer keys = NULL; 
    gpointer vals = NULL; 
    int4* index = NULL; 
    int2* group_index = NULL; 
    gsize keys_size = 0;
    gsize vals_size = 0; 
    gsize index_size = 0; 
    gsize group_index_size = 0; 
    gsize rec_num = 0;
    gsize group_num = 0;

    FETCH_BUFFER;

    g_debug("====== Rec Info ======");
    g_debug("total rec_num:%d", rec_num);
    g_debug("total keys_size:%f MB", (gdouble)keys_size/1024.0/1024.0);
    g_debug("total vals_size:%f MB", (gdouble)vals_size/1024.0/1024.0);
    g_debug("total index_size:%f MB", (gdouble)index_size/1024.0/1024.0);
}

void MarsProcessRec(Spec_t* spec, gchar type, gsize max_num, ProcessFunc_t func, gpointer arg)
{ 
    gpointer keys = NULL; 
    gpointer vals = NULL; 
    int4* index = NULL; 
    int2* group_index = NULL; 
    gsize keys_size = 0;
    gsize vals_size = 0; 
    gsize index_size = 0; 
    gsize group_index_size = 0; 
    gsize rec_num = 0;
    gsize group_num = 0;

    FETCH_BUFFER;

    int i;
    if (max_num < 0 || max_num > rec_num)
       max_num = 10;
    for (i = 0; i < max_num; i++)
    {
       gpointer key = (char*)keys + index[i].x;
       gpointer val = (char*)vals + index[i].z;
       func(key, index[i].y, val, index[i].w, i, arg); 
    }

}

void MarsProcessGroup(Spec_t* spec, gchar type, ProcessGroupFunc_t func, gpointer arg)
{
    gpointer keys = NULL; 
    gpointer vals = NULL; 
    int4* index = NULL; 
    int2* group_index = NULL; 
    gsize keys_size = 0;
    gsize vals_size = 0; 
    gsize index_size = 0; 
    gsize group_index_size = 0; 
    gsize rec_num = 0;
    gsize group_num = 0;

    FETCH_BUFFER;

    int i;
    for (i = 0; i < group_num; i++)
    {
        gpointer key = (gchar*)keys + index[group_index[i].x].x; 
        gpointer val = (gchar*)vals + index[group_index[i].x].z; 
        gsize valnum = group_index[i].y - group_index[i].x; 

        func(key, index[group_index[i].x].y, valnum, arg, spec, i);
    }
   
}

gpointer GetVal(Spec_t* spec, guint32 gindex, guint32 rindex)
{
    int4* index = &spec->_inter_index[spec->_inter_group_index[gindex].x + rindex];
    return (gchar*)spec->_inter_vals + index->z;
}
gpointer GetReduceVal(gpointer vals, int4* index, int2* group_index, guint32 gindex, guint32 rindex)
{
    int4 idx = index[group_index[gindex].x + rindex];
    return (gchar*)vals + idx.z;
}


/******
 ***/

static void mars_validate(Spec_t* spec)
{
    if (spec->workflow != MAP_ONLY &&
        spec->workflow != MAP_GROUP &&
        spec->workflow != MAP_GROUP_REDUCE)
        spec->workflow = MAP_ONLY;

    if (spec->num_map_worker <= 0)
        spec->num_map_worker = spec->_proc_num * 2;
    if (spec->num_reduce_worker <=0)
        spec->num_reduce_worker = spec->_proc_num;
}

void emit_inter_count(
                     gsize   keySize,
                     gsize   valSize,
                     MapCount_t* mc
                     )
{
    guint index = mc->index;
    //g_debug("index:%d\n", index);
    mc->key_hist[index] += keySize;
    mc->val_hist[index] += valSize;
    ++mc->index_hist[index];
}

gpointer map_count(gpointer args)
{
    MapCount_t* mc = (MapCount_t*)args;
     
    guint32 i;
    for (i = mc->start_pos; i < mc->end_pos; i++)
    {
        int4 index = mc->input_index[i];
    
        gpointer key = (char*)mc->input_keys + index.x;
        gpointer val = (char*)mc->input_vals + index.z;

        map_count_func(
                      key,
                      val,
                      index.y,
                      index.w,
                      mc
                      );
    }

    return 0;
}

void emit_inter(
               gpointer key,
               gsize   keySize,
               gpointer val,
               gsize   valSize,
               Map_t* m
               )
{
    guint32 index = m->index;

    gpointer pkey = (gpointer)((gchar*)m->output_keys + m->ps_keyoffset[index] + m->keyval_offset[index].x);
    gpointer pval = (gpointer)((gchar*)m->output_vals + m->ps_valoffset[index] + m->keyval_offset[index].y);

    memcpy(pkey, key, keySize);
    memcpy(pval, val, valSize);

    m->keyval_offset[index].x += keySize;
    m->keyval_offset[index].y += valSize;

    if (m->cur_index[index] != 0)
    {
        m->output_index[m->ps_count[index] + m->cur_index[index]].x = 
		(m->output_index[m->ps_count[index] + m->cur_index[index] - 1].x + 
		 m->output_index[m->ps_count[index] + m->cur_index[index] - 1].y);
        m->output_index[m->ps_count[index] + m->cur_index[index]].z = 
		(m->output_index[m->ps_count[index] + m->cur_index[index] - 1].z + 
		 m->output_index[m->ps_count[index] + m->cur_index[index] - 1].w);
    }	
	
    m->output_index[m->ps_count[index] + m->cur_index[index]].y = keySize;
    m->output_index[m->ps_count[index] + m->cur_index[index]].w = valSize;

    ++m->cur_index[index];
}


gpointer map(gpointer args)
{
    Map_t* m = (Map_t*)args;
    guint32 idx = m->index;

    m->output_index[m->ps_count[idx]].x = m->ps_keyoffset[idx];
    m->output_index[m->ps_count[idx]].z = m->ps_valoffset[idx];

    guint32 i;
    for (i = m->start_pos; i < m->end_pos; i++)
    {
        int4 index = m->input_index[i];
        gpointer key = (char*)m->input_keys + index.x;
        gpointer val = (char*)m->input_vals + index.z;

        map_func(
                 key,
                 val,
                 index.y,
                 index.w,
                 m
                 );
    }
}


gpointer reduce_count(gpointer args)
{
    MapCount_t* mc = (MapCount_t*)args;
     
    g_debug("start_pos:%d, end_pos:%d", mc->start_pos, mc->end_pos);

    guint32 i;
    for (i = mc->start_pos; i < mc->end_pos; i++)
    {
        int2 group_index = mc->input_group_index[i];
        int4 index = mc->input_index[group_index.x]; 
        gsize valNum = group_index.y - group_index.x;

        gpointer key = (char*)mc->input_keys + index.x;
        gpointer val = (char*)mc->input_vals;

        reduce_count_func(
                      key,
                      val,
                      index.y,
                      valNum,
                      mc,
                      i
                      );
    }

    return 0;
}
gpointer reduce(gpointer args)
{
    Map_t* m = (Map_t*)args;
    guint32 idx = m->index;

    m->output_index[m->ps_count[idx]].x = m->ps_keyoffset[idx];
    m->output_index[m->ps_count[idx]].z = m->ps_valoffset[idx];

    guint32 i;
    for (i = m->start_pos; i < m->end_pos; i++)
    {
        int2 group_index = m->input_group_index[i];
        int4 index = m->input_index[group_index.x];
        gpointer key = (char*)m->input_keys + index.x;
        gpointer val = (char*)m->input_vals;

        gsize valNum = group_index.y - group_index.x;
        reduce_func(
                 key,
                 val,
                 index.y,
                 valNum, 
                 m,
                 i
                 );
    }
}
static void mars_worker(Spec_t* spec, gchar worker_type)
{
    gpointer input_keys = NULL;
    gpointer input_vals = NULL;
    int4* input_index = NULL; 
    gsize rec_num = 0;
    gsize group_num = 0;
    guint32 num_thread = 0;

    if (worker_type == MAP_WORKER)
    {
        input_keys = spec->_input_keys;
        input_vals = spec->_input_vals;
        input_index = spec->_input_index;
        rec_num = spec->_input_num;
        num_thread = spec->num_map_worker;
    }
    else
    {
        input_keys = spec->_inter_keys;
        input_vals = spec->_inter_vals;
        input_index = spec->_inter_index;
        rec_num = spec->_inter_num;
        num_thread = spec->num_reduce_worker;
        group_num = spec->_inter_group_num;
    }

    if (rec_num <= 0)
        g_error("No input data!");

    guint32 rec_per_thread = 0;

    if (worker_type == MAP_WORKER)
    {
        rec_per_thread = rec_num / num_thread + 
            (guint32)(rec_num % num_thread != 0);
    }
    else
    {
        rec_per_thread = group_num / num_thread + 
            (guint32)(group_num % num_thread != 0);
    }

    g_debug("num_rec:%d", rec_num);
    g_debug("group_rec:%d", group_num);
    g_debug("num_thread:%d", num_thread);
    g_debug("rec_per_thread:%d",  rec_per_thread);
    g_debug("total cover:%d", rec_per_thread * num_thread);

    guint32 hist_size = sizeof(guint32) * num_thread;
    guint32* key_hist = (guint32*)g_malloc0(hist_size);
    guint32* val_hist = (guint32*)g_malloc0(hist_size);
    guint32* index_hist = (guint32*)g_malloc0(hist_size);

    //[start, end)
    guint32 i;
    gsize offset = 0;
    GThread **threads = (GThread**)g_malloc0(sizeof(GThread*)*num_thread);
    MapCount_t *mc = (MapCount_t*)g_malloc0(sizeof(MapCount_t)*num_thread);
    for (i = 0; i < num_thread; i++)
    {
        mc[i].input_keys = input_keys; 
        mc[i].input_vals = input_vals; 
        mc[i].input_index = input_index; 
        if (worker_type == REDUCE_WORKER)
            mc[i].input_group_index = spec->_inter_group_index;
        mc[i].key_hist = key_hist;
        mc[i].val_hist = val_hist;
        mc[i].index_hist = index_hist;
        mc[i].num_thread = num_thread;
        mc[i].start_pos = offset;
        offset += rec_per_thread; 

        if (worker_type == MAP_WORKER)
        {
            if (offset >= rec_num) 
               offset = rec_num;
        }
        else if (worker_type == REDUCE_WORKER)
        {
            if (offset >= group_num) 
            {   
                  offset = group_num;
                  g_debug("here");
            }
        }
        mc[i].end_pos = offset;
        mc[i].rec_per_thread = rec_per_thread;
        mc[i].index = i;

        if (worker_type == MAP_WORKER)
            threads[i] = g_thread_create(map_count, &mc[i], TRUE, NULL);    
        else
            threads[i] = g_thread_create(reduce_count, &mc[i], TRUE, NULL);
        //reduce_count(&mc[i]);
    }

    for (i = 0; i < num_thread; i++)
        g_thread_join(threads[i]);


    //scan
    guint32* ps_keyoffset = (guint32*)g_malloc0(hist_size);    
    guint32* ps_valoffset = (guint32*)g_malloc0(hist_size);
    guint32* ps_count = (guint32*)g_malloc0(hist_size);

    guint32 output_keysize = 0;
    guint32 output_valsize = 0;
    guint32 output_num = 0;
    for (i = 0; i < num_thread; i++)
    {
        ps_keyoffset[i] = output_keysize;
        ps_valoffset[i] = output_valsize;
        ps_count[i] = output_num; 

        //g_debug("output:%d,%d,%d\n", output_keysize, output_valsize, output_num);
        //g_debug("%d,%d,%d\n", key_hist[i], val_hist[i], index_hist[i]);
        output_keysize += key_hist[i];
        output_valsize += val_hist[i];
        output_num += index_hist[i];
    }

    g_free(key_hist);
    g_free(val_hist);
    g_free(index_hist);

    //map
    gpointer output_keys = (gpointer)g_malloc0(output_keysize);
    gpointer output_vals = (gpointer)g_malloc0(output_valsize);
    int4* output_index = (int4*)g_malloc0(output_num*sizeof(int4));

    int2* keyval_offset = (int2*)g_malloc0(num_thread*sizeof(int2));
    guint32*  cur_index = (guint32*)g_malloc0(num_thread*sizeof(guint32));

    Map_t *m = (Map_t*)g_malloc0(sizeof(Map_t)*num_thread);
    for (i = 0; i < num_thread; i++)
    {
        m[i].input_keys = mc[i].input_keys;
        m[i].input_vals = mc[i].input_vals;
        m[i].input_index = mc[i].input_index;
        if (worker_type == REDUCE_WORKER)
            m[i].input_group_index = spec->_inter_group_index;
        m[i].num_thread = mc[i].num_thread;
        m[i].start_pos = mc[i].start_pos;
        m[i].end_pos = mc[i].end_pos;
        m[i].rec_per_thread = mc[i].rec_per_thread;
        m[i].index = mc[i].index;

        m[i].output_keys = output_keys;
        m[i].output_vals = output_vals;
        m[i].output_index = output_index;

        m[i].ps_keyoffset = ps_keyoffset;
        m[i].ps_valoffset = ps_valoffset;
        m[i].ps_count = ps_count;
        m[i].keyval_offset = keyval_offset;
        m[i].cur_index = cur_index;

        if (worker_type == MAP_WORKER)
            threads[i] = g_thread_create(map, &m[i], TRUE, NULL);    
        else
            threads[i] = g_thread_create(reduce, &m[i], TRUE, NULL);    
    }

    for (i = 0; i < num_thread; i++)
        g_thread_join(threads[i]);

    g_free(m);
    g_free(mc);

    g_free(cur_index);
    g_free(keyval_offset);

    g_free(input_keys);
    g_free(input_vals);
    g_free(input_index);

    g_free(ps_keyoffset);
    g_free(ps_valoffset);
    g_free(ps_count);

    g_free(threads);

    if (worker_type == MAP_WORKER)
    {
        spec->_inter_keys = output_keys;
        spec->_inter_vals = output_vals;
        spec->_inter_index = output_index;
        spec->_inter_num = output_num;
        spec->_inter_keys_size = output_keysize;
        spec->_inter_vals_size = output_valsize;
        spec->_inter_index_size = output_num*sizeof(int4);
    }
    else
    {
        g_free(spec->_inter_group_index);
        spec->_output_keys = output_keys;
        spec->_output_vals = output_vals;
        spec->_output_index = output_index;
        spec->_output_num = output_num;
        spec->_output_keys_size = output_keysize;
        spec->_output_vals_size = output_valsize;
        spec->_output_index_size = output_num*sizeof(int4);
    }
}

static void mars_group(Spec_t* spec)
{
   SortGroup(spec);
}
